import unittest
import os
import pymummer
import pyfastaq
from ariba import assembly_variants, reference_data, sequence_metadata

modules_dir = os.path.dirname(os.path.abspath(assembly_variants.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestAssemblyVariants(unittest.TestCase):
    def test_get_codon_start(self):
        '''test _get_codon_start'''
        tests = [
            (0, 5, 3),
            (0, 0, 0),
            (0, 1, 0),
            (0, 2, 0),
            (1, 3, 1),
            (2, 3, 2),
            (3, 3, 3),
            (3, 6, 6),
            (3, 7, 6),
            (3, 8, 6),
        ]
        for start, position, expected in tests:
            self.assertEqual(expected, assembly_variants.AssemblyVariants._get_codon_start(start, position))


    def test_get_mummer_variants_no_variants(self):
        '''test _get_mummer_variants when no variants'''
        snp_file = os.path.join(data_dir, 'assembly_variants_test_get_mummer_variants.none.snps')
        got = assembly_variants.AssemblyVariants._get_mummer_variants(snp_file)
        self.assertEqual({}, got)


    def test_get_mummer_variants_has_variants(self):
        '''test _get_mummer_variants when there are variants'''
        snp_file = os.path.join(data_dir, 'assembly_variants_test_get_mummer_variants.snp.snps')
        v1 = pymummer.variant.Variant(pymummer.snp.Snp('42\tA\tG\t42\t42\t42\t500\t500\t1\t1\tgene\tcontig1'))
        v2 = pymummer.variant.Variant(pymummer.snp.Snp('42\tA\tG\t42\t42\t42\t500\t500\t1\t1\tgene\tcontig2'))
        v3 = pymummer.variant.Variant(pymummer.snp.Snp('40\tT\tC\t40\t42\t42\t500\t500\t1\t1\tgene\tcontig1'))
        v4 = pymummer.variant.Variant(pymummer.snp.Snp('2\tC\tG\t2\t42\t42\t500\t500\t1\t1\tgene\tcontig1'))
        expected = {
            'contig1': [[v4], [v3, v1]],
            'contig2': [[v2]]
        }
        got = assembly_variants.AssemblyVariants._get_mummer_variants(snp_file)
        self.assertEqual(expected, got)


    def test_get_variant_effect(self):
        '''test _get_variant_effect'''
        ref_seq = pyfastaq.sequences.Fasta('gene', 'GATCGCGAAGCGATGACCCATGAAGCGACCGAACGCTGA')
        v1 = pymummer.variant.Variant(pymummer.snp.Snp('6\tC\tT\t6\tx\tx\t39\t39\tx\tx\tgene\tcontig'))
        v1 = pymummer.variant.Variant(pymummer.snp.Snp('6\tC\tT\t6\tx\tx\t39\t39\tx\tx\tgene\tcontig'))
        v2 = pymummer.variant.Variant(pymummer.snp.Snp('4\tC\tA\t6\tx\tx\t39\t39\tx\tx\tgene\tcontig'))
        v3 = pymummer.variant.Variant(pymummer.snp.Snp('4\tC\tT\t4\tx\tx\t39\t39\tx\tx\tgene\tcontig'))
        v4 = pymummer.variant.Variant(pymummer.snp.Snp('6\tC\tA\t6\tx\tx\t39\t39\tx\tx\tgene\tcontig'))
        v5 = pymummer.variant.Variant(pymummer.snp.Snp('4\tC\t.\t4\tx\tx\t39\t39\tx\tx\tgene\tcontig'))
        v6 = pymummer.variant.Variant(pymummer.snp.Snp('4\t.\tA\t4\tx\tx\t39\t39\tx\tx\tgene\tcontig'))
        v7 = pymummer.variant.Variant(pymummer.snp.Snp('4\t.\tG\t4\tx\tx\t39\t39\tx\tx\tgene\tcontig'))
        v7.qry_base = 'GAT'
        v8 = pymummer.variant.Variant(pymummer.snp.Snp('4\t.\tG\t4\tx\tx\t39\t39\tx\tx\tgene\tcontig'))
        v8.qry_base = 'TGA'
        v9 = pymummer.variant.Variant(pymummer.snp.Snp('4\t.\tG\t4\tx\tx\t39\t39\tx\tx\tgene\tcontig'))
        v9.qry_base = 'ATTCCT'
        v10 = pymummer.variant.Variant(pymummer.snp.Snp('4\tC\t.\t4\tx\tx\t39\t39\tx\tx\tgene\tcontig'))
        v10.ref_base = 'CGC'
        v10.ref_end = 5
        v11 = pymummer.variant.Variant(pymummer.snp.Snp('4\tC\t.\t4\tx\tx\t39\t39\tx\tx\tgene\tcontig'))
        v11.ref_base = 'CGCGAA'
        v11.ref_end = 8

        variants = [
            ([v1], ('SYN', '.', 1)),
            ([v2], ('NONSYN', 'R2S', 1)),
            ([v2, v1], ('NONSYN', 'R2S', 1)),
            ([v3, v4], ('TRUNC', 'R2trunc', 1)),
            ([v5], ('FSHIFT', 'R2fs', 1)),
            ([v6], ('FSHIFT', 'R2fs', 1)),
            ([v7], ('INS', 'R2_E3insD', 1)),
            ([v8], ('TRUNC', 'R2trunc', 1)),
            ([v9], ('INS', 'R2_E3insIP', 1)),
            ([v10], ('DEL', 'R2del', 1)),
            ([v11], ('DEL', 'R2_E3del', 1)),
        ]

        for variant_list, expected in variants:
            self.assertEqual(expected, assembly_variants.AssemblyVariants._get_variant_effect(variant_list, ref_seq))


    def test_filter_mummer_variants(self):
        '''test filter_mummer_variants'''
        ref_seq = pyfastaq.sequences.Fasta('gene', 'GATCGCGAAGCGATGACCCATGAAGCGACCGAACGCTGA')
        v1 = pymummer.variant.Variant(pymummer.snp.Snp('4\tC\tT\t4\tx\tx\t39\t39\tx\tx\tgene\tcontig'))
        v2 = pymummer.variant.Variant(pymummer.snp.Snp('6\tC\tA\t6\tx\tx\t39\t39\tx\tx\tgene\tcontig'))
        v3 = pymummer.variant.Variant(pymummer.snp.Snp('12\tG\tT\t12\tx\tx\t39\t39\tx\tx\tgene\tcontig'))
        mummer_variants = {'contig': [[v1, v2], v3]}
        assembly_variants.AssemblyVariants._filter_mummer_variants(mummer_variants, ref_seq)
        expected = {'contig': [[v1, v2]]}
        self.assertEqual(expected, mummer_variants)


    def test_one_var_one_ctg_noncdg(self):
        '''test _get_one_variant_for_one_contig_non_coding'''
        fasta_in = os.path.join(data_dir, 'assembly_variants_one_var_one_ctg_noncdg.fa')
        tsv_in = os.path.join(data_dir, 'assembly_variants_one_var_one_ctg_noncdg.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        ref_sequence_name = 'non_coding'
        refdata_var_dict = refdata.metadata[ref_sequence_name]

        v0 = pymummer.variant.Variant(pymummer.snp.Snp('2\tT\tA\t2\tx\tx\t42\t42\tx\tx\tnon_coding\tcontig'))

        # ref has A at position 3, which is variant type. This gives contig the wild type C. Shouldn't report
        v1 = pymummer.variant.Variant(pymummer.snp.Snp('3\tA\tC\t3\tx\tx\t42\t42\tx\tx\tnon_coding\tcontig'))

        # ref has T at position 5, which is wild type. This gives contig variant type A. Should report
        v2 = pymummer.variant.Variant(pymummer.snp.Snp('5\tT\tA\t5\tx\tx\t42\t42\tx\tx\tnon_coding\tcontig'))

        meta0 = sequence_metadata.SequenceMetadata('non_coding\t0\t0\tC3A\tid1\tref has variant type A')
        meta2 = sequence_metadata.SequenceMetadata('non_coding\t0\t0\tT5A\tid1\tref has wild type T')

        mummer_variants = [v0, v1, v2]

        expected_tuples = [
            (1, 'n', 'T2A', 'SNP', [v0], set(), set()),   #0
            None,                                     #1
            (4, 'n', 'T5A', 'SNP', [v2], {meta2}, set()), #2
        ]

        expected_used_variants = [
            set(),     #0
            {meta0},   #1
            {meta2},   #2
        ]

        assert len(mummer_variants) == len(expected_tuples) == len(expected_used_variants)


        for i in range(len(mummer_variants)):
            got_tuple, got_used_variants = assembly_variants.AssemblyVariants._get_one_variant_for_one_contig_non_coding(refdata_var_dict, mummer_variants[i])
            self.assertEqual(expected_tuples[i], got_tuple)
            self.assertEqual(expected_used_variants[i], got_used_variants)


    def test_one_var_one_ctg_cdg(self):
        '''test _get_one_variant_for_one_contig_coding'''
        fasta_in = os.path.join(data_dir, 'assembly_variants_one_var_one_ctg_cdg.fa')
        tsv_in = os.path.join(data_dir, 'assembly_variants_one_var_one_ctg_cdg.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        ref_sequence_name = 'presence_absence'
        ref_sequence = refdata.sequence(ref_sequence_name)
        refdata_var_dict = refdata.metadata[ref_sequence_name]

        v0 = pymummer.variant.Variant(pymummer.snp.Snp('6\tT\tA\t6\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig'))
        v1 = pymummer.variant.Variant(pymummer.snp.Snp('9\tA\tT\t9\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig'))
        v2 = pymummer.variant.Variant(pymummer.snp.Snp('18\tG\tT\t18\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig'))
        v3 = pymummer.variant.Variant(pymummer.snp.Snp('21\tC\tT\t21\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig'))
        v4 = pymummer.variant.Variant(pymummer.snp.Snp('7\tA\tT\t7\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig'))
        v5 = pymummer.variant.Variant(pymummer.snp.Snp('12\tA\tC\t11\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig'))

        v6 = pymummer.variant.Variant(pymummer.snp.Snp('4\tG\t.\t4\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig'))
        self.assertTrue(v6.update_indel(pymummer.snp.Snp('5\tA\t.\t4\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig')))

        v7 = pymummer.variant.Variant(pymummer.snp.Snp('4\t.\tA\t4\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig'))
        self.assertTrue(v7.update_indel(pymummer.snp.Snp('4\t.\tA\t5\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig')))

        v8 = pymummer.variant.Variant(pymummer.snp.Snp('4\tG\t.\t4\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig'))
        self.assertTrue(v8.update_indel(pymummer.snp.Snp('5\tA\t.\t4\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig')))
        self.assertTrue(v8.update_indel(pymummer.snp.Snp('6\tT\t.\t4\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig')))

        v9 = pymummer.variant.Variant(pymummer.snp.Snp('4\tG\t.\t4\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig'))
        self.assertTrue(v9.update_indel(pymummer.snp.Snp('5\tA\t.\t4\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig')))
        self.assertTrue(v9.update_indel(pymummer.snp.Snp('6\tT\t.\t4\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig')))
        self.assertTrue(v9.update_indel(pymummer.snp.Snp('7\tA\t.\t4\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig')))
        self.assertTrue(v9.update_indel(pymummer.snp.Snp('8\tG\t.\t4\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig')))
        self.assertTrue(v9.update_indel(pymummer.snp.Snp('9\tA\t.\t4\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig')))

        v10 = pymummer.variant.Variant(pymummer.snp.Snp('4\t.\tA\t4\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig'))
        self.assertTrue(v10.update_indel(pymummer.snp.Snp('4\t.\tT\t5\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig')))
        self.assertTrue(v10.update_indel(pymummer.snp.Snp('4\t.\tT\t6\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig')))

        mummer_variants = [[v0], [v1], [v2], [v3], [v4], [v5], [v6], [v7], [v8], [v9], [v10]]

        meta0 = sequence_metadata.SequenceMetadata('presence_absence\t1\t0\tD2E\tid1\tref has wild type D (GAT=D, GAA=E)')
        meta4 = sequence_metadata.SequenceMetadata('presence_absence\t1\t0\tS3R\tid1\tref has variant type R (AGA=R, AGT=S)')

        expected_tuples = [
            (1, 'p', 'D2E', 'NONSYN', [v0], {meta0}, set()),    #0
            None,                                               #1
            (5, 'p', 'M6I', 'NONSYN', [v2], set(), set()),      #2
            (6, 'p', '.', 'SYN', [v3], set(), set()),           #3
            (2, 'p', 'R3trunc', 'TRUNC', [v4], set(), {meta4}), #4
            None,                                               #5
            (1, 'p', 'D2fs', 'FSHIFT', [v6], set(), {meta0}),   #6
            (1, 'p', 'D2fs', 'FSHIFT', [v7], set(), {meta0}),   #7
            (1, 'p', 'D2del', 'DEL', [v8], set(), {meta0}),     #8
            (1, 'p', 'D2_R3del', 'DEL', [v9], set(), {meta0}),  #9
            (1, 'p', 'D2_R3insI', 'INS', [v10], set(), {meta0}) #10
        ]

        expected_used_variants = [
            refdata_var_dict['p'][1], #0
            refdata_var_dict['p'][2], #1
            set(),                    #2
            set(),                    #3
            refdata_var_dict['p'][2], #4
            refdata_var_dict['p'][3], #5
            refdata_var_dict['p'][1], #6
            refdata_var_dict['p'][1], #7
            refdata_var_dict['p'][1], #8
            refdata_var_dict['p'][1], #9
            refdata_var_dict['p'][1], #10
        ]

        assert len(mummer_variants) == len(expected_tuples) == len(expected_used_variants)

        for i in range(len(mummer_variants)):
            got_tuple, got_used_variants = assembly_variants.AssemblyVariants._get_one_variant_for_one_contig_coding(ref_sequence, refdata_var_dict, mummer_variants[i])
            self.assertEqual(expected_tuples[i], got_tuple)
            self.assertEqual(expected_used_variants[i], got_used_variants)


    def test_get_remaining_known_ref_variants_amino_acids(self):
        '''test _get_remaining_known_ref_variants with amino acids'''
        ref_var1 = sequence_metadata.SequenceMetadata('gene1\t1\t0\tD2E\tid1\tfoo bar')
        ref_var2 = sequence_metadata.SequenceMetadata('gene1\t1\t0\tD3E\tid1\tfoo bar baz')
        ref_var3 = sequence_metadata.SequenceMetadata('gene1\t1\t0\tD3I\tid1\tfoo bar baz spam')
        ref_var4 = sequence_metadata.SequenceMetadata('gene1\t1\t0\tD10E\tid1\tfoo bar baz spam egg')
        ref_var5 = sequence_metadata.SequenceMetadata('gene1\t1\t0\tD14E\tid1\tfoo bar baz spam egg chips')
        ref_var6 = sequence_metadata.SequenceMetadata('gene1\t1\t0\tD15E\tid1\tfoo bar baz spam egg chips')
        ref_var7 = sequence_metadata.SequenceMetadata('gene1\t1\t0\tD40E\tid1\tfoo bar baz spam egg chips')

        known_ref_variants = {
            1: {ref_var1},
            2: {ref_var2, ref_var3},
            9: {ref_var4},
            13: {ref_var5},
            14: {ref_var6},
            39: {ref_var7}
        }

        used_ref_variants = {ref_var3, ref_var5}

        nucmer_coords = [
            pyfastaq.intervals.Interval(6, 25),
            pyfastaq.intervals.Interval(30, 100)
        ]

        expected = [(None, 'p', None, None, None, {x}, set()) for x in [ref_var2, ref_var6]]
        got = assembly_variants.AssemblyVariants._get_remaining_known_ref_variants(known_ref_variants, used_ref_variants, nucmer_coords)
        self.assertEqual(expected, got)


    def test_get_remaining_known_ref_variants_nucleotides(self):
        '''test _get_remaining_known_ref_variants with nucleotides'''
        ref_var1 = sequence_metadata.SequenceMetadata('gene1\t0\t0\tA2C\tid1\tfoo bar')
        ref_var2 = sequence_metadata.SequenceMetadata('gene1\t0\t0\tA3C\tid1\tfoo bar baz')
        ref_var3 = sequence_metadata.SequenceMetadata('gene1\t0\t0\tA3T\tid1\tfoo bar baz spam')
        ref_var4 = sequence_metadata.SequenceMetadata('gene1\t0\t0\tA10C\tid1\tfoo bar baz spam egg')
        ref_var5 = sequence_metadata.SequenceMetadata('gene1\t0\t0\tA14C\tid1\tfoo bar baz spam egg chips')
        ref_var6 = sequence_metadata.SequenceMetadata('gene1\t0\t0\tA15C\tid1\tfoo bar baz spam egg chips')
        ref_var7 = sequence_metadata.SequenceMetadata('gene1\t0\t0\tA40C\tid1\tfoo bar baz spam egg chips')

        known_ref_variants = {
            1: {ref_var1},
            2: {ref_var2, ref_var3},
            9: {ref_var4},
            13: {ref_var5},
            14: {ref_var6},
            39: {ref_var7}
        }

        used_ref_variants = {ref_var3, ref_var5}

        nucmer_coords = [
            pyfastaq.intervals.Interval(2, 13),
            pyfastaq.intervals.Interval(30, 100)
        ]

        expected = [(None, 'n', None, None, None, {x}, set()) for x in [ref_var2, ref_var4, ref_var7]]
        got = assembly_variants.AssemblyVariants._get_remaining_known_ref_variants(known_ref_variants, used_ref_variants, nucmer_coords)
        self.assertEqual(expected, got)


    def test_get_variants_presence_absence(self):
        '''test get_variants presence absence genes'''
        meta1 = sequence_metadata.SequenceMetadata('presence_absence\t1\t0\tD2E\tid1\tref has wild type D, contig has var (GAT=D, GAA=E)')
        meta2 = sequence_metadata.SequenceMetadata('presence_absence\t1\t0\tS3R\tid1\tref has variant type R, contig has wild (AGA=R, AGT=S)')
        meta3 = sequence_metadata.SequenceMetadata('presence_absence\t1\t0\tD4E\tid1\tref has variant type E, contig has var (GAA=E, GAC=D)')
        meta4 = sequence_metadata.SequenceMetadata('presence_absence\t1\t0\tA5D\tid1\tref has wild type A, contig has var (GCG=A, GAC=D)')
        meta5 = sequence_metadata.SequenceMetadata('presence_absence\t1\t0\tR13S\tid1\tref and qry have wild type')

        metadata_tsv = 'tmp.test_get_variants_presence_absence.metadata.tsv'
        with open(metadata_tsv, 'w') as f:
            print(meta1, file=f)
            print(meta2, file=f)
            print(meta3, file=f)
            print(meta4, file=f)
            print(meta5, file=f)

        fasta_in = os.path.join(data_dir, 'assembly_variants_test_get_variants_presence_absence.fa')
        refdata = reference_data.ReferenceData([fasta_in], [metadata_tsv])

        os.unlink(metadata_tsv)

        nucmer_snp_file = os.path.join(data_dir, 'assembly_variants_test_get_variants_presence_absence.snps')
        v2 = pymummer.variant.Variant(pymummer.snp.Snp('14\tC\tA\t14\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig1'))
        v3 = pymummer.variant.Variant(pymummer.snp.Snp('15\tG\tC\t15\tx\tx\t42\t42\tx\tx\tpresence_absence\tcontig1'))

        ref_nucmer_coords = {
            'contig1': [pyfastaq.intervals.Interval(0, 30)],
            'contig2': [pyfastaq.intervals.Interval(10, 41)],
        }

        ctg_nucmer_coords = {
            'contig1': [pyfastaq.intervals.Interval(0, 30)],
            'contig2': [pyfastaq.intervals.Interval(10, 41)],
        }

        expected = {
            'contig1': [
               (4, 'p', 'A5D', 'NONSYN', [v2, v3], {meta4}, set()),
               (None, 'p', None, None, None, {meta1}, set()),
               (None, 'p', None, None, None, {meta3}, set()),
            ],
            'contig2': [
               (None, 'p', None, None, None, {meta3}, set()),
               (None, 'p', None, None, None, {meta4}, set()),
               (None, 'p', None, None, None, {meta5}, set()),
            ],
        }

        a_variants = assembly_variants.AssemblyVariants(refdata, nucmer_snp_file)
        got = a_variants.get_variants('presence_absence', ctg_nucmer_coords, ref_nucmer_coords)
        self.assertEqual(expected, got)


    def test_get_variants_variants_only(self):
        '''test get_variants variants only'''
        meta1 = sequence_metadata.SequenceMetadata('variants_only\t1\t0\tD2E\tid1\tref has wild type D (GAT=D, GAA=E)')
        meta2 = sequence_metadata.SequenceMetadata('variants_only\t1\t0\tS3R\tid1\tref has variant type R (AGA=R, AGT=S)')
        meta3 = sequence_metadata.SequenceMetadata('variants_only\t1\t0\tD4E\tid1\tref has variant type E (GAA=E, GAC=D)')

        metadata_tsv = 'tmp.test_get_variants_variants_only.metadata.tsv'
        with open(metadata_tsv, 'w') as f:
            print(meta1, file=f)
            print(meta2, file=f)
            print(meta3, file=f)

        fasta_in = os.path.join(data_dir, 'assembly_variants_test_get_variants_variants_only.fa')
        refdata = reference_data.ReferenceData([fasta_in], [metadata_tsv])
        os.unlink(metadata_tsv)

        nucmer_snp_file = os.path.join(data_dir, 'assembly_variants_test_get_variants_variants_only.snps')
        v2 = pymummer.variant.Variant(pymummer.snp.Snp('14\tC\tA\t14\tx\tx\t42\t42\tx\tx\tvariants_only\tcontig1'))
        v3 = pymummer.variant.Variant(pymummer.snp.Snp('15\tG\tC\t15\tx\tx\t42\t42\tx\tx\tvariants_only\tcontig1'))

        ctg_nucmer_coords = {
            'contig1': [pyfastaq.intervals.Interval(0, 41)],
            'contig2': [pyfastaq.intervals.Interval(10, 41)],
        }

        ref_nucmer_coords = {
            'contig1': [pyfastaq.intervals.Interval(0, 41)],
            'contig2': [pyfastaq.intervals.Interval(10, 41)],
        }
        expected = {
            'contig1': [
                (4, 'p', 'A5D', 'NONSYN', [v2, v3], set(), set()),
                (None, 'p', None, None, None, {meta1}, set()),
                (None, 'p', None, None, None, {meta3}, set()),
            ],
            'contig2': [(None, 'p', None, None, None, {meta3}, set())],
        }

        a_variants = assembly_variants.AssemblyVariants(refdata, nucmer_snp_file)
        got = a_variants.get_variants('variants_only', ctg_nucmer_coords, ref_nucmer_coords)
        self.assertEqual(expected, got)




import unittest
import os
import filecmp
import pyfastaq
import pysam
from ariba import link, scaffold_graph

modules_dir = os.path.dirname(os.path.abspath(scaffold_graph.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestScaffoldGraph(unittest.TestCase):
    def test_update_from_sam(self):
        '''test update_from_sam'''
        sam_reader = pysam.Samfile(os.path.join(data_dir, 'graph_test_update_from_sam.bam'), "rb")
        ref_lengths = {}
        pyfastaq.tasks.lengths_from_fai(os.path.join(data_dir, 'graph_test_update_from_sam.ref.fa.fai'), ref_lengths)
        graph = scaffold_graph.Graph(ref_lengths)

        for sam in sam_reader.fetch(until_eof=True):
            graph.update_from_sam(sam, sam_reader)

        self.assertEqual(len(graph.partial_links), 0)
        self.assertEqual(len(graph.links), 1)
        key = ('ref1', 'ref2')
        self.assertTrue(key in graph.links)
        expected_links = [link.Link(None, None, None, s='\t'.join(['ref1', '506', 'R', '300', 'ref2', '500', 'L', '359']))]
        self.assertListEqual(graph.links[key], expected_links)


    def test_make_graph(self):
        '''test _make_graph'''
        ref_lengths = {'ref1':100, 'ref2':200, 'ref3':300}
        g = scaffold_graph.Graph(ref_lengths)

        g.partial_links = {42:42}
        with self.assertRaises(scaffold_graph.Error):
            g._make_graph(1000)
        g.partial_links = {}

        g.links[('ref1', 'ref2')] = [
            link.Link(None, None, None, '\t'.join(['ref1', '100', 'R', '50', 'ref2', '200', 'L', '10']))
        ]

        g._make_graph(10)
        self.assertEqual(len(g.contig_links), 0)

        g._make_graph(1000)
        self.assertEqual(len(g.contig_links), 1)
        expected_contig_links = {
            ('ref1', 'ref2'): {'RL': 1}
        }
        self.assertDictEqual(g.contig_links, expected_contig_links)

        g.links[('ref1', 'ref2')].append(link.Link(None, None, None, '\t'.join(['ref1', '100', 'R', '50', 'ref2', '200', 'L', '10'])))
        g.links[('ref1', 'ref2')].append(link.Link(None, None, None, '\t'.join(['ref1', '100', 'L', '50', 'ref2', '200', 'R', '10'])))
        g._make_graph(1000)
        expected_contig_links = {
            ('ref1', 'ref2'): {'RL': 2, 'LR': 1}
        }
        self.assertDictEqual(g.contig_links, expected_contig_links)


    def test_remove_low_cov_links(self):
        '''test _remove_low_cov_links'''
        ref_lengths = {'ref1':100, 'ref2':200, 'ref3':300}
        g = scaffold_graph.Graph(ref_lengths)
        g.contig_links = {
            ('ref1', 'ref2'): {'LR': 42, 'RL':41},
            ('ref2', 'ref3'): {'LR': 43, 'RL':42},
            ('ref3', 'ref4'): {'LR': 1, 'RL':2}
        }

        g._remove_low_cov_links(42)
        expected = {
            ('ref1', 'ref2'): {'LR': 42},
            ('ref2', 'ref3'): {'LR': 43, 'RL':42}
        }

        self.assertDictEqual(expected, g.contig_links)


    def test_contig_graph_is_consistent(self):
        '''test _contig_graph_is_consistent'''
        ref_lengths = {'ref1':100, 'ref2':200, 'ref3':300}
        g = scaffold_graph.Graph(ref_lengths)

        good_links = [
            {('ref1', 'ref2'): {'RL': 42}},
            {('ref1', 'ref2'): {'RL': 42}, ('ref2', 'ref3'): {'RL': 42}},
        ]

        bad_links = [
            {('ref1', 'ref2'): {'LR': 42, 'RL':42}},
            {('ref1', 'ref2'): {'LR': 42}, ('ref1', 'ref3'): {'LR': 42}}
        ]


        for d in good_links:
            g.contig_links = d
            self.assertTrue(g._contig_graph_is_consistent())

        for d in bad_links:
            g.contig_links = d
            self.assertFalse(g._contig_graph_is_consistent())


    def test_write_all_links_to_file(self):
        '''test write_all_links_to_file'''
        ref_lengths = {'ref1':100, 'ref2':200, 'ref3':300}
        g = scaffold_graph.Graph(ref_lengths)
        g.links[('ref1', 'ref2')] = [
            link.Link(None, None, None, '\t'.join(['ref1', '100', 'R', '50', 'ref2', '200', 'L', '10']))
        ]


        g.links[('ref1', 'ref2')].append(link.Link(None, None, None, '\t'.join(['ref1', '100', 'R', '50', 'ref2', '200', 'L', '10'])))
        g.links[('ref1', 'ref2')].append(link.Link(None, None, None, '\t'.join(['ref1', '100', 'L', '50', 'ref2', '200', 'R', '10'])))


        g.links[('ref3', 'ref4')] = [
            link.Link(None, None, None, '\t'.join(['ref3', '100', 'R', '42', 'ref4', '200', 'L', '42']))
        ]

        tmp_file = 'tmp.contig_links'
        g.write_all_links_to_file(tmp_file)
        self.assertTrue(filecmp.cmp(os.path.join(data_dir, 'graph_test_write_all_links_to_file.out'), tmp_file, shallow=False))
        os.unlink(tmp_file)





import os
import sys
import pyfastaq
import pymummer
import fermilite_ariba
from ariba import common, mapping, bam_parse, external_progs, ref_seq_chooser
import shlex

class Error (Exception): pass

class Assembly:
    def __init__(self,
      reads1,
      reads2,
      ref_fasta,
      ref_fastas,
      working_dir,
      final_assembly_fa,
      final_assembly_bam,
      log_fh,
      all_reference_fasta,
      contig_name_prefix='ctg',
      assembler='fermilite',
      max_insert=1000,
      min_scaff_depth=10,
      min_scaff_length=50,
      nucmer_min_id=90,
      nucmer_min_len=20,
      nucmer_breaklen=200,
      extern_progs=None,
      clean=True,
      spades_mode="wgs",
      spades_options=None,
      threads=1
    ):
        self.reads1 = os.path.abspath(reads1)
        self.reads2 = os.path.abspath(reads2)
        self.ref_fasta = os.path.abspath(ref_fasta)
        self.ref_fastas = os.path.abspath(ref_fastas)
        self.working_dir = os.path.abspath(working_dir)
        self.final_assembly_fa = os.path.abspath(final_assembly_fa)
        self.final_assembly_bam = os.path.abspath(final_assembly_bam)
        self.log_fh = log_fh
        self.all_reference_fasta = os.path.abspath(all_reference_fasta)
        self.contig_name_prefix = contig_name_prefix

        self.ref_seq_name = None
        self.assembler = assembler
        self.max_insert = max_insert
        self.min_scaff_depth = min_scaff_depth
        self.min_scaff_length = min_scaff_length
        self.nucmer_min_id = nucmer_min_id
        self.nucmer_min_len = nucmer_min_len
        self.nucmer_breaklen = nucmer_breaklen
        self.clean = clean
        self.spades_mode = spades_mode
        self.spades_options = spades_options
        self.threads = threads

        if extern_progs is None:
            self.extern_progs = external_progs.ExternalProgs(using_spades=self.assembler == 'spades')
        else:
            self.extern_progs = extern_progs

        try:
            os.mkdir(self.working_dir)
        except:
            raise Error('Error mkdir ' + self.working_dir)

        self.assembler_dir = os.path.join(self.working_dir, 'Assemble')
        self.all_assembly_contigs_fa = os.path.join(self.working_dir, 'debug_all_contigs.fa')
        self.best_assembly_fa = os.path.join(self.working_dir, 'debug_best_assembly.fa')
        self.final_assembly_fa = os.path.abspath(final_assembly_fa)


    @staticmethod
    def _run_fermilite(reads_in, fasta_out, log_out, name_prefix):
        return fermilite_ariba.fermilite_ariba(reads_in, fasta_out, log_out, name_prefix)


    def _assemble_with_fermilite(self):
        cwd = os.getcwd()
        try:
            os.chdir(self.working_dir)
        except:
            raise Error('Error chdir ' + self.working_dir)

        interleaved_reads = 'reads.fq'
        pyfastaq.tasks.interleave(self.reads1, self.reads2, interleaved_reads)
        fermilite_log = self.all_assembly_contigs_fa + '.log'
        got_from_fermilite = self._run_fermilite(interleaved_reads, self.all_assembly_contigs_fa, fermilite_log, self.contig_name_prefix)
        os.unlink(interleaved_reads)
        if os.path.exists(fermilite_log):
            with open(fermilite_log) as f:
                for line in f:
                    print(line, end='', file=self.log_fh)

            os.unlink(fermilite_log)

        self.assembled_ok = (got_from_fermilite == 0)
        os.chdir(cwd)

    @staticmethod
    def _check_spades_log_file(logfile):
        '''SPAdes can fail with a strange error. Stop everything if this happens'''
        f = pyfastaq.utils.open_file_read(logfile)

        for line in f:
            if line.startswith('== Error ==  system call for:') and line.rstrip().endswith('finished abnormally, err code: -7'):
                pyfastaq.utils.close(f)
                print('Error running SPAdes. Cannot continue. This is the error from the log file', logfile, '...', file=sys.stderr)
                print(line, file=sys.stderr)
                raise Error('Fatal error ("err code: -7") running spades. Cannot continue')

        pyfastaq.utils.close(f)
        return True

    def _assemble_with_spades(self):
        cwd = os.getcwd()
        self.assembled_ok = False
        try:
            try:
                os.chdir(self.working_dir)
            except:
                raise Error('Error chdir ' + self.working_dir)
            spades_exe = self.extern_progs.exe('spades')
            if not spades_exe:
                raise Error("Spades executable has not been found")
            spades_options = self.spades_options
            if spades_options is not None:
                spades_options = shlex.split(self.spades_options)
            if self.spades_mode == "rna":
                spades_options = ["--rna"] + (["-k","127"] if spades_options is None else spades_options)
                spades_out_seq_base = "transcripts.fasta"
            elif self.spades_mode == "sc":
                spades_options = ["--sc"] + (["-k", "33,55,77,99,127","--careful"] if spades_options is None else spades_options)
                spades_out_seq_base = "contigs.fasta"
            elif self.spades_mode == "wgs":
                spades_options = ["-k", "33,55,77,99,127","--careful"] if spades_options is None else spades_options
                spades_out_seq_base = "contigs.fasta"
            else:
                raise ValueError("Unknown spades_mode value: {}".format(self.spades_mode))
            asm_cmd = ['python3', spades_exe, "-t", str(self.threads), "--pe1-1", self.reads1, "--pe1-2", self.reads2, "-o", self.assembler_dir] + \
                spades_options
            asm_ok,err = common.syscall(asm_cmd, verbose=True, verbose_filehandle=self.log_fh, shell=False, allow_fail=True)
            if not asm_ok:
                print('Assembly finished with errors. These are the errors:', file=self.log_fh)
                print(err, file=self.log_fh)
                print('\nEnd of spades errors\n', file=self.log_fh)
            else:

                spades_log = os.path.join(self.assembler_dir, 'spades.log')
                if os.path.exists(spades_log):
                    self._check_spades_log_file(spades_log)

                    with open(spades_log) as f:
                        print('\n______________ SPAdes log ___________________\n', file=self.log_fh)
                        for line in f:
                            print(line.rstrip(), file=self.log_fh)
                        print('\n______________ End of SPAdes log _________________\n', file=self.log_fh)

                spades_warnings = os.path.join(self.assembler_dir, 'warnings.log')
                if os.path.exists(spades_warnings):
                    with open(spades_warnings) as f:
                        print('\n______________ SPAdes warnings ___________________\n', file=self.log_fh)
                        for line in f:
                            print(line.rstrip(), file=self.log_fh)
                        print('\n______________ End of SPAdes warnings _________________\n', file=self.log_fh)

                ## fermilight module generates contig names that look like `cluster_1.l15.c17.ctg.1` where 'cluster_1'==self.contig_name_prefix
                ## the whole structure of the contig name is expected in several places downstream where it is parsed into individual components.
                ## For example, it is parsed into to l and c parts in ref_seq_chooser (although the parts are not actually used).
                ## This is the code from fermilight module that generates the contig ID string:
                ## ofs << ">" << namePrefix << ".l" << overlap << ".c" << minCount << ".ctg." << i + 1 << '\n'
                ##
                ## We generate the same contig name structure here using dummy values for overlap and minCount, in order
                ## to avoid distrupting the downstream code.
                ## Note that the fermilight module generates multiple versions of the assembly on a grid of l and c values,
                ## and ref_seq_chooser then picks a single "best" (l,c) version based on coverage/identity of the nucmer
                ## alignment to the reference. Spades generates a single version of the assembly, so ref_seq_chooser
                ## can only pick that one version.

                spades_out_seq = os.path.join(self.assembler_dir,spades_out_seq_base)
                ## No need really to use general-purpose pyfastaq.sequences.file_reader here and pay performance cost for
                ## its multi-format line tests since we are only replacing the IDs in a pre-defined format
                if os.path.exists(spades_out_seq):
                    with open(spades_out_seq,"r") as inp, open(self.all_assembly_contigs_fa,"w") as out:
                        pref = self.contig_name_prefix
                        i_cont = 0
                        for line in inp:
                            if line.startswith(">"):
                                i_cont += 1
                                line = ">{}.l15.c17.ctg.{}\n".format(pref,i_cont)
                            out.write(line)
                        if i_cont > 0:
                            self.assembled_ok = True
            if self.clean:
                print('Deleting assembly directory', self.assembler_dir, file=self.log_fh)
                common.rmtree(self.assembler_dir)
        finally:
            os.chdir(cwd)


    @staticmethod
    def _fix_contig_orientation(contigs_fa, ref_fa, outfile, min_id=90, min_length=20, breaklen=200):
        '''Changes orientation of each contig to match the reference, when possible.
           Returns a set of names of contigs that had hits in both orientations to the reference'''
        if not os.path.exists(contigs_fa):
            raise Error('Cannot fix orientation of assembly contigs because file not found: ' + contigs_fa)

        tmp_coords = os.path.join(outfile + '.tmp.rename.coords')
        pymummer.nucmer.Runner(
            ref_fa,
            contigs_fa,
            tmp_coords,
            min_id=min_id,
            min_length=min_length,
            breaklen=breaklen,
            maxmatch=True,
        ).run()

        to_revcomp = set()
        not_revcomp = set()
        file_reader = pymummer.coords_file.reader(tmp_coords)
        for hit in file_reader:
            if hit.on_same_strand():
                not_revcomp.add(hit.qry_name)
            else:
                to_revcomp.add(hit.qry_name)

        os.unlink(tmp_coords)
        in_both = to_revcomp.intersection(not_revcomp)

        f = pyfastaq.utils.open_file_write(outfile)
        seq_reader = pyfastaq.sequences.file_reader(contigs_fa)
        for seq in seq_reader:
            if seq.id in to_revcomp and seq.id not in in_both:
                seq.revcomp()
            print(seq, file=f)
        pyfastaq.utils.close(f)

        return in_both


    @staticmethod
    def _parse_bam(sequences, bam, min_scaff_depth, max_insert):
        if not os.path.exists(bam):
            raise Error('File not found: ' + bam)

        bam_parser = bam_parse.Parser(bam, sequences)
        bam_parser.parse()
        bam_parser.write_files(bam)
        return bam_parser.scaff_graph_is_consistent(min_scaff_depth, max_insert)


    def run(self):
        if self.assembler == 'fermilite':
            self._assemble_with_fermilite()
        elif self.assembler == "spades":
            self._assemble_with_spades()
        print('Finished running assemblies', flush=True, file=self.log_fh)
        self.sequences = {}

        # double-check we got some contigs
        number_of_contigs = pyfastaq.tasks.count_sequences(self.all_assembly_contigs_fa) if os.path.exists(self.all_assembly_contigs_fa) else 0
        if number_of_contigs == 0:
            self.assembled_ok = False
            # This is to make this object picklable, to keep multithreading happy
            self.log_fh = None
            return
        else:
            self.assembled_ok = True

        if self.assembled_ok:
            ref_chooser = ref_seq_chooser.RefSeqChooser(
                self.ref_fastas,
                self.all_reference_fasta,
                self.all_assembly_contigs_fa,
                self.best_assembly_fa,
                self.log_fh,
                nucmer_min_id=self.nucmer_min_id,
                nucmer_min_len=self.nucmer_min_len,
                nucmer_breaklen=self.nucmer_breaklen,
            )
            ref_chooser.run()

            if ref_chooser.closest_ref_from_all_refs is None:
                print('Could not find match to reference sequences', file=self.log_fh)
                self.ref_seq_name = None
                self.log_fh = None
                return
            elif not ref_chooser.closest_ref_is_in_cluster:
                print('Closest reference', ref_chooser.closest_ref_from_all_refs, 'was not in cluster', file=self.log_fh)
                self.ref_seq_name = None
                self.log_fh = None
                return
            else:
                assert ref_chooser.closest_ref_from_all_refs is not None
                self.ref_seq_name = ref_chooser.closest_ref_from_all_refs

            print('Closest reference sequence:', self.ref_seq_name, file=self.log_fh)

            file_reader = pyfastaq.sequences.file_reader(self.ref_fastas)
            for ref_seq in file_reader:
                if self.ref_seq_name == ref_seq.id:
                    f_out = pyfastaq.utils.open_file_write(self.ref_fasta)
                    print(ref_seq, file=f_out)
                    pyfastaq.utils.close(f_out)
                    break

            contigs_both_strands = self._fix_contig_orientation(self.best_assembly_fa, self.ref_fasta, self.final_assembly_fa, min_id=self.nucmer_min_id, min_length=self.nucmer_min_len, breaklen=self.nucmer_breaklen)
            self.has_contigs_on_both_strands = len(contigs_both_strands) > 0
            pyfastaq.tasks.file_to_dict(self.final_assembly_fa, self.sequences)

            mapping.run_bowtie2(
                self.reads1,
                self.reads2,
                self.final_assembly_fa,
                self.final_assembly_bam[:-4],
                threads=self.threads,
                sort=True,
                bowtie2=self.extern_progs.exe('bowtie2'),
                bowtie2_version=self.extern_progs.version('bowtie2'),
                verbose=True,
                verbose_filehandle=self.log_fh
            )

            self.scaff_graph_ok = self._parse_bam(self.sequences, self.final_assembly_bam, self.min_scaff_depth, self.max_insert)
            print('Scaffolding graph is OK:', self.scaff_graph_ok, file=self.log_fh)

            if self.clean:
                for suffix in ['soft_clipped', 'unmapped_mates', 'scaff']:
                    filename = self.final_assembly_bam + '.' + suffix
                    print('Deleting file', filename, file=self.log_fh)
                    os.unlink(filename)


        # This is to make this object picklable, to keep multithreading happy
        self.log_fh = None



import unittest
import os
import filecmp
import pyfastaq
from ariba import samtools_variants

modules_dir = os.path.dirname(os.path.abspath(samtools_variants.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


def file2lines(filename):
    f = pyfastaq.utils.open_file_read(filename)
    lines = f.readlines()
    pyfastaq.utils.close(f)
    return lines


class TestSamtoolsVariants(unittest.TestCase):
    def test_make_vcf_and_depths_files(self):
        '''test _make_vcf_and_read_depths_files'''
        ref = os.path.join(data_dir, 'samtools_variants_make_vcf_and_depths_files.asmbly.fa')
        bam = os.path.join(data_dir, 'samtools_variants_make_vcf_and_depths_files.bam')
        expected_vcf = os.path.join(data_dir, 'samtools_variants_make_vcf_and_depths_files.expect.vcf')
        expected_depths = os.path.join(data_dir, 'samtools_variants_make_vcf_and_depths_files.expect.depths.gz')
        expected_coverage = os.path.join(data_dir, 'samtools_variants_make_vcf_and_depths_files.expect.cov')
        tmp_prefix = 'tmp.test_make_vcf_and_depths_files'
        sv = samtools_variants.SamtoolsVariants(
            ref,
            bam,
            tmp_prefix,
        )
        sv._make_vcf_and_read_depths_files()

        def get_vcf_call_lines(fname):
            with open(fname) as f:
                lines = [x for x in f.readlines() if not x.startswith('#')]
            return lines

        expected_lines = get_vcf_call_lines(expected_vcf)
        got_lines = get_vcf_call_lines(sv.vcf_file)

        # need to check that the vcf lines look the same. Column 8 is ;-delimited
        # and can be an any order.
        self.assertEqual(len(expected_lines), len(got_lines))

        for i in range(len(expected_lines)):
            expected = expected_lines[i].split('\t')
            got = got_lines[i].split('\t')
            self.assertEqual(len(expected), len(got))
            self.assertEqual(expected[:7], got[:7])
            self.assertEqual(expected[-2:], got[-2:])
            exp_set = set(expected[7].split(';'))
            got_set = set(got[7].split(';'))
            self.assertEqual(exp_set, got_set)


        # samtools-1.2 and 1.3 output not xonsistent in final column, so
        # ignore those.
        expected_lines = file2lines(expected_depths)
        got_lines = file2lines(sv.read_depths_file)
        self.assertEqual(len(expected_lines), len(got_lines))

        for i in range(len(expected_lines)):
            self.assertEqual(expected_lines[i].split('\t')[:-1], got_lines[i].split('\t')[:-1])

        os.unlink(sv.vcf_file)
        os.unlink(sv.read_depths_file)
        os.unlink(sv.read_depths_file + '.tbi')

        self.assertTrue(filecmp.cmp(expected_coverage, tmp_prefix + '.contig_depths', shallow=False))
        os.unlink(tmp_prefix + '.contig_depths')


    def test_get_read_depths(self):
        '''test _get_read_depths'''
        read_depths_file = os.path.join(data_dir, 'samtools_variants_test_get_read_depths.gz')

        tests = [
            ( ('ref1', 42), None ),
            ( ('ref2', 1), None ),
            ( ('ref1', 0), ('G', 1, '1') ),
            ( ('ref1', 2), ('T,A', 3, '2,1') ),
            ( ('ref1', 3), ('C,A,G', 42, '21,11,10') ),
            ( ('ref1', 4), ('C,AC', 41, '0,42') )
        ]

        for (name, position), expected in tests:
            self.assertEqual(expected, samtools_variants.SamtoolsVariants._get_read_depths(read_depths_file, name, position))


    def test_get_variant_positions_from_vcf(self):
        '''test _get_variant_positions_from_vcf'''
        vcf_file = os.path.join(data_dir, 'samtools_variants_test_get_variant_positions_from_vcf.vcf')

        expected = [
            ('16__cat_2_M35190.scaffold.1', 92),
            ('16__cat_2_M35190.scaffold.1', 179),
            ('16__cat_2_M35190.scaffold.1', 263),
            ('16__cat_2_M35190.scaffold.6', 93)
        ]
        self.assertEqual(expected, samtools_variants.SamtoolsVariants._get_variant_positions_from_vcf(vcf_file))


    def test_get_variants(self):
        '''test _get_variants'''
        vcf_file = os.path.join(data_dir, 'samtools_variants_test_get_variants.vcf')
        read_depths_file = os.path.join(data_dir, 'samtools_variants_test_get_variants.read_depths.gz')
        positions = [
            ('16__cat_2_M35190.scaffold.1', 92),
            ('16__cat_2_M35190.scaffold.1', 179),
            ('16__cat_2_M35190.scaffold.1', 263),
            ('16__cat_2_M35190.scaffold.6', 93)
        ]
        expected = {
            '16__cat_2_M35190.scaffold.1': {
                92: ('T,A',123, '65,58'),
                179: ('A,T', 86, '41,45'),
                263: ('G,C', 97, '53,44'),
            },
            '16__cat_2_M35190.scaffold.6': {
                93: ('T,G', 99, '56,43')
            }
        }

        got = samtools_variants.SamtoolsVariants._get_variants(vcf_file, read_depths_file, positions=positions)
        self.assertEqual(expected, got)


    def test_total_depth_per_contig(self):
        '''test total_depth_per_contig'''
        infile = os.path.join(data_dir, 'samtools_variants_test_total_depth_per_contig')
        expected = {'scaff1': 42, 'scaff64738': 11}
        got = samtools_variants.SamtoolsVariants.total_depth_per_contig(infile)
        self.assertEqual(expected, got)


    def test_variants_in_coords(self):
        '''test variants_in_coords'''
        vcf_file = os.path.join(data_dir, 'samtools_variants_test_variants_in_coords.vcf')

        nucmer_hits = {
            'scaff1': [pyfastaq.intervals.Interval(0, 41)]
        }

        expected = {'scaff1': {22}}
        got = samtools_variants.SamtoolsVariants.variants_in_coords(nucmer_hits, vcf_file)
        self.assertEqual(expected, got)


    def test_get_depths_at_position(self):
        '''test get_depths_at_position'''
        bam = os.path.join(data_dir, 'samtools_variants_test_get_depths_at_position.bam')
        ref_fa = os.path.join(data_dir, 'samtools_variants_test_get_depths_at_position.ref.fa')
        tmp_prefix = 'tmp.test_get_depths_at_position'
        samtools_vars = samtools_variants.SamtoolsVariants(
            ref_fa,
            bam,
            tmp_prefix,
        )
        samtools_vars.run()
        tests = [
            (('ref', 425), ('C,T', 31, '18,13')),
            (('not_a_ref', 10), ('ND', 'ND', 'ND')),
            (('ref', 1000000000), ('ND', 'ND', 'ND'))
        ]
        for (ref, pos), expected in tests:
            got = samtools_vars.get_depths_at_position(ref, pos)
            self.assertEqual(expected, got)

        os.unlink(samtools_vars.vcf_file)
        os.unlink(samtools_vars.read_depths_file)
        os.unlink(samtools_vars.read_depths_file + '.tbi')
        os.unlink(samtools_vars.contig_depths_file)



import pyfastaq
from ariba import mlst_profile, summary_sample


class MlstReporter:
    def __init__(self, report_tsv, mlst_file, outprefix):
        self.summary_sample = summary_sample.SummarySample(report_tsv)
        self.mlst_profile = mlst_profile.MlstProfile(mlst_file, duplicate_warnings=False)
        self.outprefix = outprefix
        self.allele_calls = {}
        self.any_allele_unsure = False


    def _call_gene(self, gene):
        results = {
            'allele': 'ND',
            'sure': None,
            'cov': '.',
            'ctgs': '.',
            'gene': None,
            'pc': '.',
            'depth': '.',
            'hetmin': '.',
            'hets': '.',
        }

        if gene not in self.summary_sample.clusters:
            return results

        cluster = self.summary_sample.clusters[gene]

        best_hit = None

        for d in cluster.data:
            if best_hit is None or best_hit['ref_base_assembled'] < d['ref_base_assembled']:
                best_hit = d

        assert best_hit is not None

        het_data = []

        for d in cluster.data:
            if d['ctg'] == best_hit['ctg']:
                if d['var_type'] == 'HET':
                    het_data.append(d['smtls_nts_depth'])
                    depths = [int(x) for x in d['smtls_nts_depth'].split(',')]
                    depths.sort()
                    het_pc = round(100.0 * depths[-1] / sum(depths), 2)
                    if results['hetmin'] == '.' or results['hetmin'] > het_pc:
                        results['hetmin'] = het_pc
        if len(het_data):
            results['hets'] = '.'.join(het_data)

        results['cov'] = round(100.0 * best_hit['ref_base_assembled'] / best_hit['ref_len'], 2)
        results['ctgs'] = len({x['ctg'] for x in cluster.data})
        results['depth'] = best_hit['ctg_cov']
        results['pc'] = best_hit['pc_ident']
        results['allele'] = int(best_hit['ref_name'].split('.')[-1])
        results['sure'] = True

        if results['hetmin'] != '.' or results['cov'] < 100 or results['ctgs'] != 1 or results['pc'] < 100:
            results['sure'] = False

        return results


    def _call_genes(self):
        self.gene_results = {gene: self._call_gene(gene) for gene in self.mlst_profile.genes_list}


    def _call_sequence_type(self):
        type_dict = {gene: self.gene_results[gene]['allele'] for gene in self.mlst_profile.genes_list if self.gene_results[gene]['allele'] is not None}
        self.sequence_type = self.mlst_profile.get_sequence_type(type_dict)


    def _report_strings(self, results):
        allele_str = str(results['allele'])
        if results['sure'] is not None and not results['sure']:
            self.any_allele_unsure = True
            allele_str += '*'

        details_list = [str(results[x]) for x in ['cov', 'pc', 'ctgs', 'depth', 'hetmin', 'hets']]
        return allele_str, '\t'.join(details_list)


    def _write_reports(self):
        f_out_all = pyfastaq.utils.open_file_write(self.outprefix + '.details.tsv')
        print('gene\tallele\tcov\tpc\tctgs\tdepth\thetmin\thets', file=f_out_all)
        out_simple = [str(self.sequence_type)]

        for gene in self.mlst_profile.genes_list:
            allele_str, detail_str = self._report_strings(self.gene_results[gene])
            out_simple.append(allele_str)
            print(gene, allele_str, detail_str, sep='\t', file=f_out_all)

        pyfastaq.utils.close(f_out_all)

        if self.sequence_type != 'ND' and self.any_allele_unsure:
            out_simple[0] += '*'

        f_out_simple = pyfastaq.utils.open_file_write(self.outprefix + '.tsv')
        print('ST', *self.mlst_profile.genes_list, sep='\t', file=f_out_simple)
        print(*out_simple, sep='\t', file=f_out_simple)
        pyfastaq.utils.close(f_out_simple)


    def run(self):
        self.summary_sample.run()
        self._call_genes()
        self._call_sequence_type()
        self._write_reports()




import unittest
import filecmp
import os
from ariba import vfdb_parser

modules_dir = os.path.dirname(os.path.abspath(vfdb_parser.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestVfdbParser(unittest.TestCase):
    def test_fa_header_to_name_pieces(self):
        '''test _fa_header_to_name_pieces'''
        tests = [
            ('VF123(gi:1234) (abcD) foobar description [abc] [genus species]', ('VF123(gi:1234)', 'abcD', 'foobar description [abc]', 'genus species')),
            ('F123(gi:1234) (abcD) foobar description [abc] [genus species]', None), # no V at start
            ('VF123(gi:1234) (abcD) foobar description [abc]', None),  # missing genus species
            ('VF123(gi:1234) abcD foobar description [abc] [genus species]', None), # brackets missing around abcD
        ]

        for header, expected in tests:
            got = vfdb_parser.VfdbParser._fa_header_to_name_pieces(header)
            self.assertEqual(expected, got)


    def test_fa_header_to_name_and_metadata(self):
        '''test _fa_header_to_name_and_metadata'''
        headers = [
            'VF123(gi:1234) (abcD) foobar description [abc] [genus species]',
            'F123(gi:1234) (abcD) foobar description [abc] [genus species]', # no V at start
            'VF123(gi:1234) (abcD) foobar description [abc]', # missing genus species
            'VF123(gi:1234) abcD foobar description [abc] [genus species]', # brackets missing around abcD
        ]

        expected = [
            ('abcD.VF123(gi:1234).genus_species', 'foobar description [abc]'),
            (headers[1], '.'),
            (headers[2], '.'),
            (headers[3], '.'),
        ]

        assert len(headers) == len(expected)
        for i in range(len(headers)):
            got = vfdb_parser.VfdbParser._fa_header_to_name_and_metadata(headers[i])
            self.assertEqual(expected[i], got)


    def test_run(self):
        '''test run'''
        infile = os.path.join(data_dir, 'vfdb_parser_test_run.in.fa')
        expected_tsv = os.path.join(data_dir, 'vfdb_parser_test_run.out.tsv')
        expected_fa = os.path.join(data_dir, 'vfdb_parser_test_run.out.fa')
        outprefix = 'tmp.vfdb_parser_test_run'
        got_tsv = outprefix + '.tsv'
        got_fa = outprefix + '.fa'
        vp = vfdb_parser.VfdbParser(infile, outprefix)
        vp.run()
        self.assertTrue(filecmp.cmp(expected_tsv, got_tsv, shallow=False))
        self.assertTrue(filecmp.cmp(expected_fa, got_fa, shallow=False))
        os.unlink(got_tsv)
        os.unlink(got_fa)



import unittest
import os
import pyfastaq
import pysam
from ariba import bam_parse, link

modules_dir = os.path.dirname(os.path.abspath(bam_parse.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


def file_to_lines(filename):
    f = pyfastaq.utils.open_file_read(filename)
    lines = [l.rstrip() for l in f]
    pyfastaq.utils.close(f)
    return lines


class TestBamParse(unittest.TestCase):
    def test_sam_to_soft_clipped(self):
        '''test _sam_to_soft_clipped'''
        bam = os.path.join(data_dir, 'bam_parse_test_sam_to_soft_clipped.bam')
        bp = bam_parse.Parser(bam, {'x':pyfastaq.sequences.Fasta('x', 'ACGT')})
        sam_reader = pysam.Samfile(bam, "rb")
        i = 0
        expected = [
            None,
            (False, False),
            (True, False),
            (False, True),
            (True, True)
        ]

        for sam in sam_reader.fetch(until_eof=True):
            if i == 0:
                with self.assertRaises(bam_parse.Error):
                    bp._sam_to_soft_clipped(sam)
            else:
                self.assertEqual(expected[i], bp._sam_to_soft_clipped(sam))

            i += 1

    def test_update_soft_clipped_from_sam(self):
        '''test _update_soft_clipped_from_sam'''
        ref_seqs = {}
        ref_fasta = os.path.join(data_dir, 'bam_parse_test_update_soft_clipped_from_sam.ref.fa')
        bam = os.path.join(data_dir, 'bam_parse_test_update_soft_clipped_from_sam.bam')
        pyfastaq.tasks.file_to_dict(ref_fasta, ref_seqs)
        bp = bam_parse.Parser(bam, ref_seqs)
        expected = [
            {},
            {'ref1': {61: [1, 0]}},
            {'ref1': {61: [2, 0], 118: [0, 1]}},
        ]
        i = 0
        sam_reader = pysam.Samfile(bam, "rb")
        for sam in sam_reader.fetch(until_eof=True):
            bp._update_soft_clipped_from_sam(sam)
            self.assertEqual(expected[i], bp.soft_clipped)
            i += 1


    def test_update_unmapped_mates_from_sam(self):
        '''test _update_unmapped_mates_from_sam'''
        ref_seqs = {}
        ref_fasta = os.path.join(data_dir, 'bam_parse_test_update_unmapped_mates_from_sam.ref.fa')
        bam = os.path.join(data_dir, 'bam_parse_test_update_unmapped_mates_from_sam.bam')
        pyfastaq.tasks.file_to_dict(ref_fasta, ref_seqs)
        bp = bam_parse.Parser(bam, ref_seqs)
        expected = [
            {},
            {'ref': {240: 1}},
            {'ref': {240: 1, 360: 1}},
            {'ref': {240: 1, 360: 1}},
            {'ref': {240: 1, 360: 1}},
            {'ref': {240: 1, 360: 1}},
            {'ref': {240: 1, 360: 1}},
            {'ref': {240: 1, 360: 1}},
        ]
        i = 0
        sam_reader = pysam.Samfile(bam, "rb")
        for sam in sam_reader.fetch(until_eof=True):
            bp._update_unmapped_mates_from_sam(sam)
            self.assertEqual(expected[i], bp.unmapped_mates)
            i += 1


    def test_parse(self):
        '''test parse'''
        ref_seqs = {}
        ref_fasta = os.path.join(data_dir, 'bam_parse_test_parse.ref.fa')
        bam = os.path.join(data_dir, 'bam_parse_test_parse.bam')
        pyfastaq.tasks.file_to_dict(ref_fasta, ref_seqs)
        bp = bam_parse.Parser(bam, ref_seqs)
        bp.parse()
        expected_soft_clipped = {'ref1': {2: [1, 0], 58: [0, 1]}}
        expected_unmapped_mates = {'ref1': {2: 1}}
        l = link.Link(None, None, None, s='\t'.join(['ref1', '500', 'R', '240', 'ref2', '500', 'L', '299']))
        expected_link_keys = [('ref1', 'ref2')]
        self.assertEqual(expected_soft_clipped, bp.soft_clipped)
        self.assertEqual(expected_unmapped_mates, bp.unmapped_mates)
        self.assertEqual(list(bp.scaff_graph.links.keys()), expected_link_keys)
        self.assertEqual(len(bp.scaff_graph.links[expected_link_keys[0]]), 1)
        self.assertEqual(bp.scaff_graph.links[expected_link_keys[0]][0], l)


    def test_write_soft_clipped_to_file(self):
        '''test _write_soft_clipped_to_file'''
        tmp_file = 'tmp.soft_clipped'
        bam = os.path.join(data_dir, 'dummy.bam')
        bp = bam_parse.Parser(bam, {'x':pyfastaq.sequences.Fasta('x', 'ACGT')})
        soft_clipped = {'ref1': {42: [142, 242], 43: [44, 45]}}
        bp.soft_clipped = soft_clipped
        bp._write_soft_clipped_to_file(tmp_file)
        expected_lines = [
            'ref1\t43\t142\t242',
            'ref1\t44\t44\t45'
        ]
        got_lines = file_to_lines(tmp_file)
        self.assertEqual(got_lines, expected_lines)
        os.unlink(tmp_file)


    def test_write_unmapped_mates_to_file(self):
        '''test _write_unmapped_mates_to_file'''
        tmp_file = 'tmp.unmapped_mates'
        bam = os.path.join(data_dir, 'dummy.bam')
        bp = bam_parse.Parser(bam, {'x':pyfastaq.sequences.Fasta('x', 'ACGT')})
        bp.unmapped_mates = {'ref': {41: 1, 42: 43}}
        bp._write_unmapped_mates_to_file(tmp_file)
        expected_lines = ['ref\t42\t1', 'ref\t43\t43']
        got_lines = file_to_lines(tmp_file)
        self.assertEqual(got_lines, expected_lines)
        os.unlink(tmp_file)



import unittest
import json
import shutil
import os
import pickle
import pyfastaq
import filecmp
from ariba import clusters, common, external_progs, histogram, sequence_metadata

modules_dir = os.path.dirname(os.path.abspath(clusters.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')
extern_progs = external_progs.ExternalProgs()


def file_to_list(infile):
    f = pyfastaq.utils.open_file_read(infile)
    lines = [x for x in f.readlines()]
    pyfastaq.utils.close(f)
    return lines


class TestClusters(unittest.TestCase):
    def setUp(self):
        self.cluster_dir = 'tmp.Cluster'
        self.refdata_dir = 'tmp.RefData'
        os.mkdir(self.refdata_dir)
        shutil.copyfile(os.path.join(data_dir, 'clusters_test_dummy_db.fa'), os.path.join(self.refdata_dir, '02.cdhit.all.fa'))
        shutil.copyfile(os.path.join(data_dir, 'clusters_test_dummy_db.tsv'), os.path.join(self.refdata_dir, '01.filter.check_metadata.tsv'))
        with open(os.path.join(self.refdata_dir, '00.info.txt'), 'w') as f:
            print('genetic_code\t11', file=f)

        with open(os.path.join(self.refdata_dir, '02.cdhit.clusters.pickle'), 'wb') as f:
            pickle.dump({'0': {'x'}}, f)

        reads1 = os.path.join(data_dir, 'clusters_test_dummy_reads_1.fq')
        reads2 = os.path.join(data_dir, 'clusters_test_dummy_reads_2.fq')
        self.clusters = clusters.Clusters(self.refdata_dir, reads1, reads2, self.cluster_dir, extern_progs, clean=False)


    def tearDown(self):
        common.rmtree(self.cluster_dir)
        common.rmtree(self.refdata_dir)


    def test_load_reference_data_info_file(self):
        '''test _load_reference_data_info_file'''
        infile = os.path.join(data_dir, 'clusters_test_load_data_info_file')
        expected = {'genetic_code': 11}
        got = clusters.Clusters._load_reference_data_info_file(infile)
        self.assertEqual(expected, got)


    def test_load_ref_data_from_dir(self):
        '''test _load_reference_data_from_dir'''
        indir = os.path.join(data_dir, 'clusters_load_ref_data_from_dir')
        got_refdata, got_clusters = clusters.Clusters._load_reference_data_from_dir(indir)
        expected_seq_dict = {
            'variants_only1': pyfastaq.sequences.Fasta('variants_only1', 'atggcgtgcgatgaataa'),
            'presabs1': pyfastaq.sequences.Fasta('presabs1', 'atgatgatgagcccggcgatggaaggcggctag'),
            'noncoding1': pyfastaq.sequences.Fasta('noncoding1', 'ACGTA'),
        }
        self.assertEqual(expected_seq_dict, got_refdata.sequences)
        self.assertEqual(11, got_refdata.genetic_code)

        expected_metadata = {
            'presabs1': {
                'seq_type': 'p',
                'variant_only': False,
                '.': {sequence_metadata.SequenceMetadata('presabs1\t1\t0\t.\t.\tpresabs1 description')},
                'n': {},
                'p': {}
            },
            'variants_only1': {
                'seq_type': 'p',
                'variant_only': True,
                '.': set(),
                'n': {},
                'p': {1: {sequence_metadata.SequenceMetadata('variants_only1\t1\t1\tC2I\t.\tdescription of variants_only1 C2I')}}
            },
            'noncoding1': {
                'seq_type': 'n',
                'variant_only': False,
                '.': {sequence_metadata.SequenceMetadata('noncoding1\t0\t0\t.\t.\t.')},
                'n': {},
                'p': {},
            }
        }
        self.assertEqual(expected_metadata, got_refdata.metadata)

        expected_clusters = {'0': {'presabs1'}, '1': {'variants_only1'}, '2': {'noncoding1'}}
        self.assertEqual(expected_clusters, got_clusters)

        self.assertEqual({'a': 'b'}, got_refdata.extra_parameters)

    def test_minimap_reads_to_all_ref_seqs(self):
        '''test test_minimap_reads_to_all_ref_seqs'''
        clusters_tsv = os.path.join(data_dir, 'clusters_minimap_reads_to_all_refs.clstrs.tsv')
        ref_fasta = os.path.join(data_dir, 'clusters_minimap_reads_to_all_refs.ref.fa')
        reads_1 = os.path.join(data_dir, 'clusters_minimap_reads_to_all_refs.reads_1.fq')
        reads_2 = os.path.join(data_dir, 'clusters_minimap_reads_to_all_refs.reads_2.fq')
        tmp_outprefix = 'tmp.clusters_test_minimap_reads_to_all_ref_seqs'
        clusters.Clusters._minimap_reads_to_all_ref_seqs(clusters_tsv, ref_fasta, reads_1, reads_2, tmp_outprefix)
        expected_cluster2rep = os.path.join(data_dir, 'clusters_minimap_reads_to_all_refs.out.clstr2rep')
        expected_cluster_counts = os.path.join(data_dir, 'clusters_minimap_reads_to_all_refs.out.clstr_count')
        expected_proper_pairs = os.path.join(data_dir, 'clusters_minimap_reads_to_all_refs.out.pairs')
        expected_insert_hist = os.path.join(data_dir, 'clusters_minimap_reads_to_all_refs.out.hist')

        # not sure that the reads order is preserved, so just check read store file exists
        self.assertTrue(os.path.exists(os.path.join(tmp_outprefix + '.reads')))

        self.assertTrue(filecmp.cmp(expected_cluster2rep, tmp_outprefix + '.cluster2representative', shallow=False))
        self.assertTrue(filecmp.cmp(expected_cluster_counts, tmp_outprefix + '.clusterCounts', shallow=False))
        self.assertTrue(filecmp.cmp(expected_proper_pairs, tmp_outprefix + '.properPairs', shallow=False))
        self.assertTrue(filecmp.cmp(expected_insert_hist, tmp_outprefix + '.insertHistogram', shallow=False))
        os.unlink(tmp_outprefix + '.cluster2representative')
        os.unlink(tmp_outprefix + '.clusterCounts')
        os.unlink(tmp_outprefix + '.properPairs')
        os.unlink(tmp_outprefix + '.insertHistogram')
        os.unlink(tmp_outprefix + '.reads')


    def test_load_minimap_out_cluster2representative(self):
        '''test _load_minimap_out_cluster2representative'''
        infile = os.path.join(data_dir, 'clusters_test_load_minimap_out_cluster2representative.in')
        got = clusters.Clusters._load_minimap_out_cluster2representative(infile)
        expected = {'1': 'ref2', '2': 'ref42'}
        self.assertEqual(expected, got)


    def test_load_minimap_out_cluster_counts(self):
        '''test _load_minimap_out_cluster_counts'''
        infile = os.path.join(data_dir, 'clusters_test_load_minimap_out_cluster_counts.in')
        got_read_count, got_base_count = clusters.Clusters._load_minimap_out_cluster_counts(infile)
        expected_read_count = {'1': 42, '2': 43}
        expected_base_count = {'1': 4242, '2': 4343}
        self.assertEqual(got_read_count, expected_read_count)
        self.assertEqual(got_base_count, expected_base_count)


    def test_load_minimap_insert_histogram(self):
        '''test _load_minimap_insert_histogram'''
        infile = os.path.join(data_dir, 'clusters_test_load_minimap_insert_histogram.in')
        bin_size = 10
        got = clusters.Clusters._load_minimap_insert_histogram(infile, bin_size)
        expected = histogram.Histogram(bin_size)
        expected.add(85, count=1)
        expected.add(86, count=2)
        expected.add(90, count=4)
        expected.add(91, count=6)
        expected.add(97, count=10)
        expected.add(100, count=7)
        expected.add(111, count=3)
        self.assertEqual(expected, got)


    def test_load_minimap_proper_pairs(self):
        '''test _load_minimap_proper_pairs'''
        infile = os.path.join(data_dir, 'clusters_test_load_minimap_proper_pairs.in')
        got = clusters.Clusters._load_minimap_proper_pairs(infile)
        self.assertEqual(42424242, got)


    def test_load_minimap_files(self):
        '''test _load_minimap_files'''
        bin_size = 10
        inprefix = os.path.join(data_dir, 'clusters_test_load_minimap_files')
        got_clster2rep, got_cluster_read_count, got_cluster_base_count, got_insert_hist, got_proper_pairs = clusters.Clusters._load_minimap_files(inprefix, bin_size)
        expected_clster2rep = {'1': 'ref2', '2': 'ref42'}
        expected_cluster_read_count = {'1': 42, '2': 43}
        expected_cluster_base_count = {'1': 4242, '2': 4343}
        expected_insert_hist_bins = {80: 3, 90: 20, 100: 7, 110: 3}
        expected_proper_pairs = 42424242
        self.assertEqual(expected_clster2rep, got_clster2rep)
        self.assertEqual(expected_cluster_read_count, got_cluster_read_count)
        self.assertEqual(expected_cluster_base_count, got_cluster_base_count)
        self.assertEqual(expected_insert_hist_bins, got_insert_hist.bins)
        self.assertEqual(expected_proper_pairs, got_proper_pairs)


    def test_set_insert_size_data(self):
        '''test _set_insert_size_data'''
        self.clusters.insert_hist.bins = {
            1: 1,
            2: 1,
            3: 3,
            4: 3,
            5: 5,
            6: 3,
            7: 2,
            8: 2,
            9: 1,
            10: 1,
        }
        self.clusters.insert_hist.bin_width=1

        self.clusters._set_insert_size_data()
        self.assertEqual(self.clusters.insert_size, 5.5)
        self.assertEqual(self.clusters.insert_sspace_sd, 0.91)


    def test_write_report(self):
        class FakeCluster:
            def __init__(self, lines):
                self.report_lines = lines

        clusters_dict = {
            'gene1': FakeCluster(['gene1\tline1']),
            'gene2': FakeCluster(['gene2\tline2'])
        }

        tmp_tsv = 'tmp.test_write_report.tsv'
        clusters.Clusters._write_report(clusters_dict, tmp_tsv)

        expected = os.path.join(data_dir, 'clusters_test_write_report.tsv')
        self.assertTrue(filecmp.cmp(expected, tmp_tsv, shallow=False))
        os.unlink(tmp_tsv)


    def test_write_catted_assemblies_fasta(self):
        '''test _write_catted_assemblies_fasta'''
        seq1 = pyfastaq.sequences.Fasta('seq1', 'ACGT')
        seq2 = pyfastaq.sequences.Fasta('seq2', 'TTTT')
        seq3 = pyfastaq.sequences.Fasta('seq3', 'AAAA')
        class FakeAssembly:
            def __init__(self, seqs):
                if seqs is not None:
                    self.sequences = {x.id: x for x in seqs}

        class FakeCluster:
            def __init__(self, seqs):
                self.assembly = FakeAssembly(seqs)

        self.clusters.clusters = {
            'cluster1': FakeCluster([seq1, seq2]),
            'cluster2': FakeCluster([seq3]),
            'cluster3': FakeCluster(None),
        }

        tmp_file = 'tmp.test_write_catted_assemblies_fasta.fa'
        self.clusters._write_catted_assemblies_fasta(tmp_file)
        expected = os.path.join(data_dir, 'clusters_test_write_catted_assemblies_fasta.expected.out.fa')
        self.assertTrue(filecmp.cmp(expected, tmp_file, shallow=False))
        os.unlink(tmp_file)


    def test_write_catted_assembled_seqs_fasta(self):
        '''test _write_catted_assembled_seqs_fasta'''
        seq1 = pyfastaq.sequences.Fasta('seq1', 'ACGT')
        seq2 = pyfastaq.sequences.Fasta('seq2', 'TTTT')
        seq3 = pyfastaq.sequences.Fasta('seq3', 'AAAA')
        class FakeAssemblyCompare:
            def __init__(self, assembled_seqs):
                if assembled_seqs is not None:
                    self.assembled_reference_sequences = {x.id: x for x in assembled_seqs}

        class FakeCluster:
            def __init__(self, assembled_seqs):
                self.assembly_compare = FakeAssemblyCompare(assembled_seqs)

        self.clusters.clusters = {
            'gene1': FakeCluster([seq1, seq2]),
            'gene2': FakeCluster([seq3]),
            'gene3': FakeCluster(None),
        }

        tmp_file = 'tmp.test_write_catted_assembled_seqs_fasta.fa'
        self.clusters._write_catted_assembled_seqs_fasta(tmp_file)
        expected = os.path.join(data_dir, 'clusters_test_write_catted_assembled_genes_fasta.expected.out.fa')
        self.assertTrue(filecmp.cmp(expected, tmp_file, shallow=False))
        os.unlink(tmp_file)


    def test_cat_genes_match_ref(self):
        '''test _write_catted_genes_matching_refs_fasta'''
        seq1 = pyfastaq.sequences.Fasta('seq1', 'ACGT')
        seq3 = pyfastaq.sequences.Fasta('seq3', 'AAAA')
        class FakeAssemblyCompare:
            def __init__(self, seq, seq_type, start, end):
                self.gene_matching_ref = seq
                self.gene_matching_ref_type = seq_type
                self.gene_start_bases_added = start
                self.gene_end_bases_added = end

        class FakeCluster:
            def __init__(self, seq, seq_type, start, end):
                self.assembly_compare = FakeAssemblyCompare(seq, seq_type, start, end)

        self.clusters.clusters = {
            'gene1': FakeCluster(seq1, 'TYPE1', 1, 3),
            'gene2': FakeCluster(None, None, None, None),
            'gene3': FakeCluster(seq3, 'TYPE3', 4, 5),
        }

        tmp_file = 'tmp.test_write_catted_genes_matching_refs_fasta.fa'
        self.clusters._write_catted_genes_matching_refs_fasta(tmp_file)
        expected = os.path.join(data_dir, 'clusters_cat_genes_match_ref.fa')
        self.assertTrue(filecmp.cmp(expected, tmp_file, shallow=False))
        os.unlink(tmp_file)


    def test_write_mlst_reports(self):
        '''test _write_mlst_reports'''
        ariba_report = os.path.join(data_dir, 'clusters_test_write_mlst_reports.ariba.report.tsv')
        mlst_file = os.path.join(data_dir, 'clusters_test_write_mlst_reports.mlst_profile.tsv')
        expected_short = os.path.join(data_dir, 'clusters_test_write_mlst_reports.out.tsv')
        expected_long = os.path.join(data_dir, 'clusters_test_write_mlst_reports.out.details.tsv')
        outprefix = 'tmp.test_clusters__write_mlst_reports'
        got_short = outprefix + '.tsv'
        got_long = outprefix + '.details.tsv'

        clusters.Clusters._write_mlst_reports('filenotthere', ariba_report, outprefix)
        self.assertFalse(os.path.exists(got_short))
        self.assertFalse(os.path.exists(got_long))

        clusters.Clusters._write_mlst_reports(mlst_file, ariba_report, outprefix)
        self.assertTrue(filecmp.cmp(expected_short, got_short, shallow=False))
        self.assertTrue(filecmp.cmp(expected_long, got_long, shallow=False))
        os.unlink(got_short)
        os.unlink(got_long)


    def test_write_tb_resistance_calls_json(self):
        '''test _write_tb_resistance_calls_json'''
        ariba_report = os.path.join(data_dir, 'clusters_write_tb_resistance_calls_json.in.tsv')
        tmp_out = 'tmp.write_tb_resistance_calls_json.out.json'
        clusters.Clusters._write_tb_resistance_calls_json(ariba_report, tmp_out)
        expected = os.path.join(data_dir, 'clusters_write_tb_resistance_calls_json.out.json')
        self.assertTrue(filecmp.cmp(expected, tmp_out, shallow=False))
        os.unlink(tmp_out)


    def test_run_with_tb(self):
        '''test complete run with TB amr calling'''
        tmp_out = 'tmp.clusters_run_with_tb'
        c = clusters.Clusters(
            os.path.join(data_dir, 'clusters_run_with_tb.ref'),
            os.path.join(data_dir, 'clusters_run_with_tb.reads_1.fq.gz'),
            os.path.join(data_dir, 'clusters_run_with_tb.reads_2.fq.gz'),
            tmp_out,
            extern_progs,
        )
        c.run()

        expect_json = {"Pyrazinamide": [[ "pncA", "A3E"]]}
        json_file = os.path.join(tmp_out, 'tb.resistance.json')
        self.assertTrue(os.path.exists(json_file))
        with open(json_file) as f:
            got_json = json.load(f)
        self.assertEqual(expect_json, got_json)
        common.rmtree(tmp_out)



import unittest
import os
import re
from ariba import cdhit, external_progs


modules_dir = os.path.dirname(os.path.abspath(cdhit.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')
extern_progs = external_progs.ExternalProgs()

class TestCdhit(unittest.TestCase):
    def test_init_fail_infile_missing(self):
        '''test init_fail_infile_missing'''
        with self.assertRaises(cdhit.Error):
            cdhit.Runner('oopsnotafile', 'out')


    def test_init_fail_invalid_memory(self):
        '''test_init_fail_invalid_memory'''
        infile = os.path.join(data_dir, 'cdhit_test_run.in.fa')
        with self.assertRaises(cdhit.Error):
            cdhit.Runner(infile, memory_limit=-10)


    def test_get_clusters_from_bak_file(self):
        '''test _get_clusters_from_bak_file'''
        infile = os.path.join(data_dir, 'cdhit_test_get_clusters_from_bak_file.in')
        expected = {
            '0': {'seq1', 'seq2', 'seq3'},
            '1': {'seq4'},
            '2': {'seq5'}
        }
        got = cdhit.Runner._get_clusters_from_bak_file(infile)
        self.assertEqual(expected, got)

        expected = {
            '42': {'seq1', 'seq2', 'seq3'},
            '43': {'seq4'},
            '44': {'seq5'}
        }
        got = cdhit.Runner._get_clusters_from_bak_file(infile, min_cluster_number=42)
        self.assertEqual(expected, got)


    def test_run(self):
        '''test run'''
        infile = os.path.join(data_dir, 'cdhit_test_run.in.fa')
        r = cdhit.Runner(infile)
        clusters = r.run()
        expected_clusters = {
            '0': {'seq1', 'seq2', 'seq3'},
            '1': {'seq4'},
        }
        self.assertEqual(clusters, expected_clusters)


    def test_run_min_cluster_number_42(self):
        '''test run with min_cluster_number 42'''
        infile = os.path.join(data_dir, 'cdhit_test_run.in.fa')
        r = cdhit.Runner(infile, min_cluster_number=42)
        clusters = r.run()
        expected_clusters = {
            '42': {'seq1', 'seq2', 'seq3'},
            '43': {'seq4'},
        }
        self.assertEqual(clusters, expected_clusters)


    def test_fake_run(self):
        '''test fake_run'''
        infile = os.path.join(data_dir, 'cdhit_test_fake_run.in.fa')
        r = cdhit.Runner(infile)
        clusters = r.fake_run()
        expected_clusters = {
            '0': {'seq1'},
            '1': {'seq2'},
            '2': {'seq3'},
            '3': {'seq4'},
        }
        self.assertEqual(clusters, expected_clusters)


    def test_fake_run_cluster_min_42(self):
        '''test fake_run min_cluster 42'''
        infile = os.path.join(data_dir, 'cdhit_test_fake_run.in.fa')
        r = cdhit.Runner(infile, min_cluster_number=42)
        clusters = r.fake_run()
        expected_clusters = {
            '42': {'seq1'},
            '43': {'seq2'},
            '44': {'seq3'},
            '45': {'seq4'},
        }
        self.assertEqual(clusters, expected_clusters)


    def test_fake_run_fail(self):
        '''test fake_run with non-unique names'''
        infile = os.path.join(data_dir, 'cdhit_test_fake_run.non-unique.in.fa')
        r = cdhit.Runner(infile)
        with self.assertRaises(cdhit.Error):
            r.fake_run()


    def test_load_user_clusters_file_good_file(self):
        '''test _load_user_clusters_file with good input file'''
        infile = os.path.join(data_dir, 'cdhit_test_load_user_clusters_file.good')
        expected  = {
            '0': {'seq1', 'seq2', 'seq3'},
            '1': {'seq4'},
            '2': {'seq5', 'seq6'}
        }

        got = cdhit.Runner._load_user_clusters_file(infile, {'seq' + str(i) for i in range(1,7,1)})
        self.assertEqual(expected, got)

        expected['2'] = {'seq5'}
        got = cdhit.Runner._load_user_clusters_file(infile, {'seq' + str(i) for i in range(1,6,1)})
        self.assertEqual(expected, got)


    def test_load_user_clusters_file_good_file_with_renaming(self):
        '''test _load_user_clusters_file with good input file with some renamed'''
        rename_dict = {'seq2': 'seq2_renamed', 'seq6': 'seq6_renamed'}
        infile = os.path.join(data_dir, 'cdhit_test_load_user_clusters_file.good')
        expected  = {
            '0': {'seq1', 'seq2_renamed', 'seq3'},
            '1': {'seq4'},
            '2': {'seq5', 'seq6_renamed'}
        }

        names = {'seq1', 'seq2_renamed', 'seq3', 'seq4', 'seq5', 'seq6_renamed'}
        got = cdhit.Runner._load_user_clusters_file(infile, names, rename_dict=rename_dict)
        self.assertEqual(expected, got)


    def test_load_user_clusters_file_bad_file(self):
        '''test _load_user_clusters_file with bad input files'''
        infiles = [
            os.path.join(data_dir, 'cdhit_test_load_user_clusters_file.bad1'),
            os.path.join(data_dir, 'cdhit_test_load_user_clusters_file.bad2'),
            os.path.join(data_dir, 'cdhit_test_load_user_clusters_file.bad3')
        ]
        for filename in infiles:
            with self.assertRaises(cdhit.Error):
                cdhit.Runner._load_user_clusters_file(filename, {'seq1', 'seq2', 'seq3'})


    def test_run_get_clusters_from_file(self):
        '''test run_get_clusters_from_file'''
        fa_infile = os.path.join(data_dir, 'cdhit_test_run_get_clusters_from_dict.in.fa')
        clusters_infile = os.path.join(data_dir, 'cdhit_test_run_get_clusters_from_dict.in.clusters')
        r = cdhit.Runner(fa_infile)
        clusters = r.run_get_clusters_from_file(clusters_infile, {'seq1', 'seq2', 'seq3'})
        expected_clusters = {
            '0': {'seq1', 'seq2'},
            '1': {'seq3'},
        }
        self.assertEqual(clusters, expected_clusters)


    def test_run_get_clusters_from_file_with_renaming(self):
        '''test run_get_clusters_from_file with renaming'''
        rename_dict = {'seq2': 'seq2_renamed'}
        fa_infile = os.path.join(data_dir, 'cdhit_test_run_get_clusters_from_dict_rename.in.fa')
        clusters_infile = os.path.join(data_dir, 'cdhit_test_run_get_clusters_from_dict.in.clusters')
        r = cdhit.Runner(fa_infile)
        clusters = r.run_get_clusters_from_file(clusters_infile, {'seq1', 'seq2_renamed', 'seq3'}, rename_dict=rename_dict)
        expected_clusters = {
            '0': {'seq1', 'seq2_renamed'},
            '1': {'seq3'},
        }
        self.assertEqual(clusters, expected_clusters)


    def test_get_run_cmd_with_default_memory(self):
        '''test_get_run_cmd_with_default_memory'''
        fa_infile = os.path.join(data_dir, 'cdhit_test_run_get_clusters_from_dict_rename.in.fa')
        r = cdhit.Runner(fa_infile)
        run_cmd = r.get_run_cmd('foo/bar/file.out')
        match = re.search('^.+ -o foo/bar/file.out -c 0.9 -T 1 -s 0.0 -d 0 -bak 1$', run_cmd)
        self.assertIsNotNone(match, msg="Command output was " + run_cmd)


    def test_get_run_cmd_with_non_default_memory(self):
        '''test_get_run_cmd_with_non_default_memory'''
        fa_infile = os.path.join(data_dir, 'cdhit_test_run_get_clusters_from_dict_rename.in.fa')
        r = cdhit.Runner(fa_infile, memory_limit=900)
        run_cmd = r.get_run_cmd('foo/bar/file.out')
        match = re.search('^.+ -o foo/bar/file.out -c 0.9 -T 1 -s 0.0 -d 0 -bak 1 -M 900$', run_cmd)
        self.assertIsNotNone(match, msg="Command output was " + run_cmd)


    def test_get_run_cmd_with_unlimited_memory(self):
        '''test_get_run_cmd_with_unlimited_memory'''
        fa_infile = os.path.join(data_dir, 'cdhit_test_run_get_clusters_from_dict_rename.in.fa')
        r = cdhit.Runner(fa_infile, memory_limit=0)
        run_cmd = r.get_run_cmd('foo/bar/file.out')
        match = re.search('^.+ -o foo/bar/file.out -c 0.9 -T 1 -s 0.0 -d 0 -bak 1 -M 0$', run_cmd)
        self.assertIsNotNone(match, msg="Command output was " + run_cmd)



import urllib.request
from bs4 import BeautifulSoup
from distutils.version import LooseVersion


class Error (Exception): pass


class MegaresDataFinder:
    def __init__(self, version=None):
        self.url_root = 'https://megares.meglab.org/download/'
        self.index_url = self.url_root + 'index.php'
        self.version = version


    def _get_available_zips(self):
        try:
            response = urllib.request.urlopen(self.index_url)
            html_text = response.read()
        except:
            raise Error('Error getting megares download page ' + self.index_url)

        return html_text


    @classmethod
    def _zips_from_index_page_string(cls, html_text):
        try:
            soup = BeautifulSoup(html_text, 'html.parser')
        except:
            raise Error('Error parsing contents of megares download page. Cannot continue')

        prefix = 'megares_v'
        suffix = '.zip'
        zips = {}

        for link in soup.find_all('a'):
            href = link.get('href')
            if href.startswith(prefix) and href.endswith(suffix):
                version = href[len(prefix):-len(suffix)]
                zips[version] = href

        return zips


    @classmethod
    def _get_url_for_version(cls, zips, version=None):
        if version is None:
            versions = list(zips.keys())
            versions.sort(key=LooseVersion)
            return zips[versions[-1]]
        else:
            try:
                return zips[version]
            except:
                versions = ', '.join(list(zips.keys()))
                raise Error('Error! version ' + version + ' of megares not found. Available versions: ' + versions)


    def run(self):
        print('Finding available megares versions from', self.index_url)
        html_text = self._get_available_zips()
        zips = MegaresDataFinder._zips_from_index_page_string(html_text)
        print('Found versions: ', ', '.join(list(zips.keys())))
        url = MegaresDataFinder._get_url_for_version(zips, version=self.version)
        return self.url_root + url





import shutil
import sys
import os
import pyfastaq
from ariba import mlst_profile, pubmlst_getter, ref_preparer, versions

class Error (Exception): pass


class PubmlstRefPreparer:
    def __init__(self, species, outdir, debug=False, verbose=False):
        self.species = species
        self.outdir = outdir
        self.debug = debug
        self.verbose = verbose
        self.clusters_file = os.path.join(outdir, 'clusters.tsv')
        self.mlst_download_dir = os.path.join(self.outdir, 'pubmlst_download')
        self.prepareref_dir = os.path.join(self.outdir, 'ref_db')
        self.extern_progs, version_report_lines = versions.get_all_versions()


    @classmethod
    def _filter_seq_dict(cls, d):
        lengths = [len(x) for x in d.values()]
        lengths.sort()
        median_length = lengths[int(len(d) / 2)]
        min_length = 0.9 * median_length
        max_length = 1.1 * median_length
        to_delete = []
        for seq in d:
            if not min_length <= len(d[seq]) <= max_length:
                to_delete.append(seq)
                print('WARNING: Median sequence length is', median_length, 'but', seq, 'has length', len(d[seq]), 'which is too long or short. Removing.', file=sys.stderr)

        for seq in to_delete:
            del d[seq]


    def _load_fasta_files_and_write_clusters_file(self, indir):
        self.sequences = {}
        self.fasta_files = []
        clusters_fh = pyfastaq.utils.open_file_write(self.clusters_file)

        for gene_name in self.profile.genes_list:
            infile = os.path.join(indir, gene_name + '.tfa')

            if not os.path.exists(infile):
                pyfastaq.utils.close(clusters_fh)
                raise Error('Cannot find file "' + infile + '" for gene ' + gene_name)

            self.sequences[gene_name] = {}
            pyfastaq.tasks.file_to_dict(infile, self.sequences[gene_name])
            PubmlstRefPreparer._filter_seq_dict(self.sequences[gene_name])
            seq_names = sorted(list(self.sequences[gene_name].keys()))
            print(*seq_names, sep='\t', file=clusters_fh)

            if self.verbose:
                print('Loaded fasta file for gene', gene_name)

            with open(infile, 'w') as f:
                for seq in sorted(self.sequences[gene_name]):
                    print(self.sequences[gene_name][seq], file=f)

            self.fasta_files.append(infile)

        pyfastaq.utils.close(clusters_fh)


    def run(self):
        try:
            os.mkdir(self.outdir)
        except:
            raise Error('Error making output directory ' + self.outdir)

        pubmlst = pubmlst_getter.PubmlstGetter(debug=self.debug, verbose=self.verbose)
        pubmlst.get_species_files(self.species, self.mlst_download_dir)
        if self.verbose:
            print('Downloaded data from pubmlst')

        profile_file = os.path.join(self.mlst_download_dir, 'profile.txt')
        self.profile = mlst_profile.MlstProfile(profile_file, duplicate_warnings=True)
        if self.verbose:
            print('Loaded mlst profile file', profile_file)

        self._load_fasta_files_and_write_clusters_file(self.mlst_download_dir)
        if self.verbose:
            print('Loaded fasta files and wrote clusters file')
            print('Putting data in ariba db directory', self.prepareref_dir)

        refprep = ref_preparer.RefPreparer(
            self.fasta_files,
            self.extern_progs,
            all_coding='no',
            clusters_file=self.clusters_file,
            verbose=self.verbose,
        )
        refprep.run(self.prepareref_dir)
        shutil.copy(profile_file, os.path.join(self.prepareref_dir, 'pubmlst.profile.txt'))

        print('ariba db directory prepared. You can use it like this:')
        print('ariba run', self.prepareref_dir, 'reads_1.fq reads_2.fq output_directory')
